import re
from typing import Iterable, Any, Type, Union

from pydantic import BaseModel

from assimilator.core.database import BaseRepository, Specification
from assimilator.core.database.exceptions import NotFoundError
from assimilator.internal.database.specifications import DictSpecification, DictKeySpecification


class DictModel(BaseModel):
    id: str
    value: Any


class InternalRepository(BaseRepository):
    def __init__(self, session: dict, model: Type[DictModel] = DictModel):
        super(InternalRepository, self).__init__(session)
        self.model = model

    def get_initial_query(self):
        return ''

    def get(self, *specifications: DictSpecification, lazy: bool = False):
        try:
            return self.session[self.apply_specifications(specifications)]
        except (KeyError, TypeError) as exc:
            raise NotFoundError(exc)

    def filter(self, *specifications: DictSpecification, lazy: bool = False):
        models = []
        key_mask = self.apply_specifications(specifications)

        for key, value in self.session.items():
            if re.match(key, key_mask):
                models.append(value)

        return models

    def save(self, obj: DictModel):
        obj_copy = obj.copy()
        self.session[obj_copy.id] = obj_copy

    def update_many(self, specifications: Iterable[Specification], updated_fields: dict):
        keys = self.apply_specifications(specifications)

        for key in self.session:
            if re.match(key, keys):
                for field, value in updated_fields.items():
                    setattr(self.session[key].value, field, value)

    def delete(self, obj):
        del self.session[obj.id]

    def update(self, obj: DictModel):
        obj_copy = obj.copy()
        self.session[obj_copy.id] = obj_copy

    def is_modified(self, obj):
        return self.get(DictKeySpecification(obj.id)) == obj

    def refresh(self, obj):
        obj.value = self.get(DictKeySpecification(obj.id))



import json
from typing import Type

from pydantic import ValidationError, BaseModel

from assimilator.core.database.exceptions import ParsingError


class JSONParsedMixin:
    @classmethod
    def from_json(cls: Type['BaseModel'], data: str):
        try:
            return cls(**json.loads(data))
        except ValidationError as exc:
            raise ParsingError(exc)



from assimilator.core.database import Specification


class DictSpecification(Specification):
    def apply(self, query: str) -> str:     # returns the key
        return super(DictSpecification, self).apply(query)


class DictKeySpecification(DictSpecification):
    def __init__(self, id: str):
        self.id = id

    def apply(self, query):
        return f"{query}{self.id}"



from assimilator.internal.events.events_bus import *
from assimilator.internal.database.repository import *
from assimilator.internal.database.unit_of_work import *
from assimilator.internal.database.specifications import *



from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field

from assimilator.core.patterns.mixins import JSONParsedMixin


class Event(JSONParsedMixin, BaseModel):
    id: int
    event_name: str
    event_date: datetime = Field(default_factory=datetime.now)


class UnknownEvent(Event):
    """ The event type is unknown, so all the fields are in data """
    data: Any


class AckEvent(Event):
    acknowledged: bool = False



from assimilator.alchemy.database.repository import AlchemyRepository


class AlchemyOutboxRepository(AlchemyRepository):
    def __init__(self, event_model, session, initial_query=None):
        super(AlchemyOutboxRepository, self).__init__(session, initial_query)
        self.event_model = event_model

    def save(self, obj):
        super(AlchemyOutboxRepository, self).save(obj)
        super(AlchemyOutboxRepository, self).save(self.event_model(obj.outbox_event))



from assimilator.core.database.exceptions import ParsingError


class EventError(Exception):
    pass


class EventParsingError(ParsingError, EventError):
    pass


class EventProducingError(EventError):
    pass



from assimilator.core.database import UnitOfWork
from assimilator.core.database.repository import BaseRepository


class DictUnitOfWork(UnitOfWork):
    def __init__(self, repository: BaseRepository):
        super(DictUnitOfWork, self).__init__(repository)
        self._saved_data = None

    def begin(self):
        self._saved_data = self.repository.session
        self.repository.session = dict(self._saved_data)

    def rollback(self):
        self.repository.session = self._saved_data

    def commit(self):
        self._saved_data = self.repository.session

    def close(self):
        pass



from abc import ABC, abstractmethod
from typing import Iterable

from assimilator.core.database.specifications import Specification


class BaseRepository(ABC):
    def __init__(self, session):
        self.session = session

    @abstractmethod
    def get_initial_query(self):
        raise NotImplementedError("get_initial_query() must be implemented")

    def apply_specifications(self, specifications):
        query = self.get_initial_query()

        for specification in specifications:
            query = specification(query)

        return query

    @abstractmethod
    def get(self, *specifications: Specification, lazy: bool = False):
        raise NotImplementedError("get() is not implemented()")

    @abstractmethod
    def filter(self, *specifications: Specification, lazy: bool = False):
        raise NotImplementedError("filter() is not implemented()")

    @abstractmethod
    def save(self, obj):
        raise NotImplementedError("save() is not implemented in the repository")

    @abstractmethod
    def update_many(self, specifications: Iterable[Specification], updated_fields: dict):
        raise NotImplementedError("update_many() is not implemented in the repository")

    @abstractmethod
    def delete(self, obj):
        raise NotImplementedError("delete() is not implemented in the repository")

    @abstractmethod
    def update(self, obj):
        raise NotImplementedError("update() is not implemented in the repository")

    @abstractmethod
    def is_modified(self, obj):
        raise NotImplementedError("is_modified() is not implemented in the repository")

    @abstractmethod
    def refresh(self, obj):
        raise NotImplementedError("refresh() is not implemented in the repository")



from abc import ABC, abstractmethod
from functools import wraps


class Specification(ABC):
    @abstractmethod
    def apply(self, query):
        raise NotImplementedError("Specification must specify apply()")

    def __call__(self, query):
        return self.apply(query)


def specification(func: callable, *args, **kwargs):
    @wraps(func)
    def specification_wrapper(query):
        return func(query, *args, **kwargs)

    return specification_wrapper
